/*
AC3 Filter for Cool Edit/Adobe Audition
Copyright  2007 James Chapman
http://www.vuplayer.com

Utilizes the liba52 library
Copyright  2000-2002 Michel Lespinasse <walken@zoy.org>
Copyright  1999-2000 Aaron Holtzman <aholtzma@ess.engr.uvic.ca>
http://liba52.sourceforge.net/
*/

#include <windows.h>
#include <stdio.h>
#include "../filters.h"
#include "Cac3.h"

#define CHUNKSIZE			65536

//////////////////////////////////////////////////////////////////////
// REQUIRED COOL EDIT FILTER FUNCTIONS
//////////////////////////////////////////////////////////////////////

__declspec(dllexport) short FAR PASCAL QueryCoolFilter(COOLQUERY* cq)
{
	ZeroMemory(cq,sizeof(COOLQUERY));
	strcpy(cq->szName,"AC3");		
	strcpy(cq->szCopyright,"Copyright  liba52.sourceforge.net");
	strcpy(cq->szExt,"ac3");
	cq->lChunkSize=CHUNKSIZE;
	cq->dwFlags=QF_CANLOAD|QF_CANDO32BITFLOATS;
	cq->Stereo32=0xFF;
 	return C_VALIDLIBRARY;
}

__declspec(dllexport) BOOL FAR PASCAL FilterUnderstandsFormat(LPSTR filename)
{
	if(strlen(filename)>3&&!_stricmp(filename+strlen(filename)-3,"ac3"))
	{
		Cac3 ac3;
		if(ac3.Open(filename))
			return 1;
	}
	return 0;
}

//////////////////////////////////////////////////////////////////////
// REQUIRED COOL EDIT FILTER FUNCTIONS - READING
//////////////////////////////////////////////////////////////////////

__declspec(dllexport) DWORD FAR PASCAL FilterGetFileSize(HANDLE hInput)
{
	Cac3* ac3=(Cac3*)hInput;
	if(ac3->m_totalsamples*8>0xfffffff8)
		return 0xfffffff8;
	else
		return (DWORD)(ac3->m_totalsamples*8);
}

__declspec(dllexport) HANDLE FAR PASCAL OpenFilterInput(LPSTR filename,long* lRate,WORD* wBPS,WORD* wChannels,HWND hWnd,long* lChunkSize)
{
	Cac3* ac3=new Cac3();
	if(ac3)
	{
		if(ac3->Open(filename,32768.0f))
		{
			*lRate=ac3->m_samplerate;
			*wBPS=32;
			*wChannels=2;
			*lChunkSize=CHUNKSIZE;
			return ac3;
		}
		delete ac3;
	}
	return 0;
}


__declspec(dllexport) DWORD FAR PASCAL ReadFilterInput(HANDLE hInput, unsigned char* buf, long lBytes)
{
	Cac3* ac3=(Cac3*)hInput;
	return ac3->Read((float*)buf,lBytes);
}

__declspec(dllexport) void FAR PASCAL CloseFilterInput(HANDLE hInput)
{
	Cac3* ac3=(Cac3*)hInput;
	if(ac3)
		delete ac3;
}

//////////////////////////////////////////////////////////////////////
// OPTIONAL COOL EDIT FILTER FUNCTIONS
//////////////////////////////////////////////////////////////////////

__declspec(dllexport) DWORD FAR PASCAL FilterOptionsString(HANDLE hInput, LPSTR lpszString)
{
	Cac3* ac3=(Cac3*)hInput;
	if(ac3)
	{
		char buf[32]="Channels: ";
		switch(ac3->m_flags&0x0f)
		{
		case A52_MONO:
			strcat(buf,"1");
			break;
		case A52_STEREO:
		case A52_DOLBY:
			strcat(buf,"2");
			break;
		case A52_3F:
		case A52_2F1R:
			strcat(buf,"3");
			break;
		case A52_3F1R:
		case A52_2F2R:
			strcat(buf,"4");
			break;
		case A52_3F2R:
			strcat(buf,"5");
			break;
		default:
			buf[0]=0;
		}
		if((ac3->m_flags&A52_LFE)&&strlen(buf))
			strcat(buf,".1");
		if(strlen(buf))
		{
			strcat(buf,"\nBitrate: ");
			_itoa(ac3->m_bitrate/1000,buf+strlen(buf),10);
			strcat(buf,"kbps");
		}
		strcpy(lpszString,buf);
	}
	return 0;
}
