/*
Ogg Vorbis File Filter for Cool Edit/Adobe Audition
Copyright  2007 James Chapman
http://www.vuplayer.com

Ogg Vorbis is Copyright  1994-2007 the Xiphophorus Company
http://www.vorbis.com
*/

#include <windows.h>
#include <commctrl.h>
#include <time.h>
#include <stdlib.h>
#include "../filters.h"
#include "resource.h"

#include <vorbis/codec.h>
#include <vorbis/vorbisfile.h>
#include <vorbis/vorbisenc.h>

#define CHUNKSIZE			65536
#define MAX_VBR				1100
#define MAX_ABR				511
#define DEFAULT_VBR			500
#define DEFAULT_ABR			128

#define TAGCOUNT			15

#define FCC_INFO			"INFO"
#define FCC_WAVE			"WAVE"
#define FCC_DISP			"DISP"

typedef struct
{
	LPCSTR riff;
	LPCSTR tag;
} TAGDATA;

const TAGDATA TAGS[TAGCOUNT]=
{
	{"IART","ARTIST"},
	{"ICMT","COMMENT"},
	{"ICOP","COPYRIGHT"},
	{"ICRD","DATE"},
	{"IENG","ENGINEER"},
	{"IGNR","GENRE"},
	{"IKEY","KEYWORDS"},
	{"IMED","MEDIUM"},
	{"INAM","ALBUM"},
	{"ISFT","SOFTWARE"},
	{"ISRC","SOURCE"},
	{"ITCH","TECHNICIAN"},
	{"ISBJ","SUBJECT"},
	{"ISRF","SOURCE FORM"},
	{"DISP","TITLE"}
};

typedef struct	// any special vars associated with output file
{
	WORD channels;
	WORD bps;
	BOOL skipfirst;
	ogg_stream_state os;
	ogg_page og;
	ogg_packet op;
	vorbis_info vi;
	vorbis_comment vc;
	vorbis_dsp_state vd;
	vorbis_block vb;
	FILE* hFile;
} FILTER_OUTPUT;

typedef struct	// any special vars associated with input file
{
	DWORD options;
	WORD channels;
	OggVorbis_File vfile;
	float** vfloat;
	long vpos;
	long vsize;
} FILTER_INPUT;

void EnableOptions(HWND hWndDlg, BOOL vbr);
BOOL GetNextComment(vorbis_comment* comment, SPECIALDATA* psp);
BOOL ANSItoUTF8(char* ansi, char** utf8);
BOOL UTF8toANSI(char* utf8, char** ansi);
void EnsureCRLF(WCHAR** str);

//////////////////////////////////////////////////////////////////////
// REQUIRED COOL EDIT FILTER FUNCTIONS
//////////////////////////////////////////////////////////////////////

__declspec(dllexport) short FAR PASCAL QueryCoolFilter(COOLQUERY* cq)
{
	ZeroMemory(cq,sizeof(COOLQUERY));
	strcpy(cq->szName,"Ogg Vorbis");		
	strcpy(cq->szCopyright,"Ogg Vorbis 1.2.0");
	strcpy(cq->szExt,"ogg");
	cq->lChunkSize=CHUNKSIZE; 
	cq->dwFlags=QF_CANLOAD|QF_CANSAVE|QF_RATEADJUSTABLE|QF_HASOPTIONSBOX|QF_CANDO32BITFLOATS|QF_READSPECIALFIRST|QF_WRITESPECIALFIRST;
	cq->Stereo8=0xFF;
	cq->Stereo16=0xFF;
	cq->Stereo32=0xFF;
	cq->Mono8=0xFF;
	cq->Mono16=0xFF;
	cq->Mono32=0xFF;
 	return C_VALIDLIBRARY;
}

__declspec(dllexport) BOOL FAR PASCAL FilterUnderstandsFormat(LPSTR filename)
{
	OggVorbis_File vf;
	if(!ov_fopen(filename,&vf))
	{	
		ov_clear(&vf);
		return TRUE;
	}
	return FALSE;	
}

//////////////////////////////////////////////////////////////////////
// REQUIRED COOL EDIT FILTER FUNCTIONS - READING
//////////////////////////////////////////////////////////////////////

__declspec(dllexport) HANDLE FAR PASCAL OpenFilterInput(LPSTR filename,long* lRate,WORD* wBPS,WORD* wChannels,HWND hWnd,long* lChunkSize)
{
	FILTER_INPUT *filter=new FILTER_INPUT;
	if(filter)
	{
		ZeroMemory(filter,sizeof(FILTER_INPUT));
		if(!ov_fopen(filename,&filter->vfile))
		{
			vorbis_info* vi=ov_info(&filter->vfile,-1);
			filter->channels=vi->channels;
			*lRate=vi->rate;
			*wBPS=32;
			*wChannels=vi->channels;
			*lChunkSize=CHUNKSIZE;
			return filter;
		}
		delete filter;
	}
	return 0;
}

__declspec(dllexport) DWORD FAR PASCAL FilterGetFileSize(HANDLE hInput)
{
	FILTER_INPUT *filter=(FILTER_INPUT*)hInput;
	__int64 size=ov_pcm_total(&filter->vfile,-1)*filter->channels*4;
	if(size>0xffffffff)
	{
		size=0x100000000;
		size-=filter->channels*4;
	}
	return (DWORD)size;
}

__declspec(dllexport) DWORD FAR PASCAL ReadFilterInput(HANDLE hInput, unsigned char* buf, long lBytes)
{
	FILTER_INPUT *filter=(FILTER_INPUT*)hInput;
	float* f=(float*)buf;
	int bitstream;
	int i;
	long pos=0;
	while(pos<lBytes/4)
	{
		if(filter->vpos<filter->vsize)
		{
			for(i=0;i<filter->channels;i++)
				f[pos++]=*(filter->vfloat[i]+filter->vpos)*32768.0f;
			filter->vpos++;
		}
		else
		{
			filter->vpos=0;
			i=4;
			while(--i)
			{
				filter->vsize=ov_read_float(&filter->vfile,&filter->vfloat,CHUNKSIZE/4,&bitstream);
				if(filter->vsize>=0)
					break;
			}
			if(!i||!filter->vsize)
				break;
		}
	}
	return pos*4;
}

__declspec(dllexport) void FAR PASCAL CloseFilterInput(HANDLE hInput)
{
	FILTER_INPUT *filter=(FILTER_INPUT*)hInput;
	if(filter)
	{
		ov_clear(&filter->vfile);
		delete filter;
	}
}

//////////////////////////////////////////////////////////////////////
// REQUIRED COOL EDIT FILTER FUNCTIONS - WRITING
//////////////////////////////////////////////////////////////////////

__declspec(dllexport) void FAR PASCAL GetSuggestedSampleType(long* lRate, WORD* wBPS, WORD *wChannels)
{
	*wBPS=0;
	*lRate=0;
	*wChannels=0;
}

__declspec(dllexport) HANDLE FAR PASCAL OpenFilterOutput(LPSTR filename,
														 long samprate,
														 WORD BPS,
														 WORD channels,
														 long lSize,
														 long* lChunkSize,
														 DWORD options)
{
	FILTER_OUTPUT *filter=new FILTER_OUTPUT;
	if(!filter)
		return 0;
	ZeroMemory(filter,sizeof(FILTER_OUTPUT));
	filter->channels=channels;
	filter->bps=BPS;
	*lChunkSize=CHUNKSIZE/8;
	BOOL vbr=1;
	int bitrate=0;
	int minimum=-1;
	int maximum=-1;
	float quality=0;
	if((options>>31)&1)			// Use variable bitrate encoding
	{
		bitrate=(options&0xffff);
		if(bitrate>MAX_VBR)
			bitrate=DEFAULT_VBR;
		quality=(float)(bitrate-100);
		quality/=1000;
	}
	else							
	{
		if((options>>30)&1)		// Use managed bitrate encoding
		{
			vbr=0;
			bitrate=(int)(options&0x1ff);
			if(!bitrate)
				bitrate=DEFAULT_ABR;
			if((options>>29)&1)
			{
				minimum=(int)((options>>9)&0x1ff);
				if(minimum>bitrate)
					minimum=-1;
				else
					minimum*=1000;
			}
			if((options>>28)&1)
			{
				maximum=(int)((options>>18)&0x1ff);
				if(maximum<bitrate)
					maximum=-1;
				else
					maximum*=1000;
			}
			bitrate*=1000;
		}
		else
		{
			quality=(float)(DEFAULT_VBR-100);
			quality/=1000;
		}
	}
	vorbis_info_init(&filter->vi);
	int result;
	if(vbr)
		result=vorbis_encode_init_vbr(&filter->vi,channels,samprate,quality);
	else
		result=vorbis_encode_init(&filter->vi,channels,samprate,maximum,bitrate,minimum);
	if(result)
	{
		delete filter;
		return 0;
	}
	FILE* hFile=fopen(filename,"wb");
	if(hFile)
		filter->hFile=hFile;
	else
	{
		delete filter;
		return 0;
	}
	vorbis_comment_init(&filter->vc);
	return filter;
}

__declspec(dllexport) DWORD FAR PASCAL WriteFilterOutput(HANDLE hOutput, BYTE* pBuf, long lBytes)
{
	FILTER_OUTPUT *filter=(FILTER_OUTPUT*)hOutput;
	if(!filter->skipfirst)
	{
		filter->skipfirst=1;
		vorbis_analysis_init(&filter->vd,&filter->vi);
		vorbis_block_init(&filter->vd,&filter->vb);
		srand(time(0));
		ogg_stream_init(&filter->os,rand());
		ogg_packet header;
		ogg_packet header_comm;
		ogg_packet header_code;
		vorbis_analysis_headerout(&filter->vd,&filter->vc,&header,&header_comm,&header_code);
		ogg_stream_packetin(&filter->os,&header);
		ogg_stream_packetin(&filter->os,&header_comm);
		ogg_stream_packetin(&filter->os,&header_code);
		int result=ogg_stream_flush(&filter->os,&filter->og);
		while(result)
		{
			fwrite(filter->og.header,1,filter->og.header_len,filter->hFile);
			fwrite(filter->og.body,1,filter->og.body_len,filter->hFile);
			result=ogg_stream_flush(&filter->os,&filter->og);
		}
	}
	float **buffer=vorbis_analysis_buffer(&filter->vd,lBytes/filter->channels/(filter->bps/8));
	long i,j;
	switch(filter->bps)
	{
	case 8:
		{
			for(i=0;i<(lBytes/filter->channels);i++)
			{
				for(j=0;j<filter->channels;j++)
					buffer[j][i]=((char)(pBuf[i*filter->channels+j]+128))/128.0f;
			}
		}
		break;
	case 16:
		{
			short* buf=(short*)pBuf;
			for(i=0;i<(lBytes/filter->channels/2);i++)
			{
				for(j=0;j<filter->channels;j++)
					buffer[j][i]=buf[i*filter->channels+j]/32768.0f;
			}
		}
		break;
	case 32:
		{
			float* buf=(float*)pBuf;
			for(i=0;i<(lBytes/filter->channels/4);i++)
			{
				for(j=0;j<filter->channels;j++)
					buffer[j][i]=buf[i*filter->channels+j]/32768.0f;
			}
		}
	}
	int result=0;
	vorbis_analysis_wrote(&filter->vd,i);
	while(vorbis_analysis_blockout(&filter->vd,&filter->vb)==1)
	{
		vorbis_analysis(&filter->vb,0);
		vorbis_bitrate_addblock(&filter->vb);
		while(vorbis_bitrate_flushpacket(&filter->vd,&filter->op))
		{
			ogg_stream_packetin(&filter->os,&filter->op);
			result=ogg_stream_pageout(&filter->os,&filter->og);
			while(result)
			{
				fwrite(filter->og.header,1,filter->og.header_len,filter->hFile);
				fwrite(filter->og.body,1,filter->og.body_len,filter->hFile);
				if(ogg_page_eos(&filter->og))
					break;
				result=ogg_stream_pageout(&filter->os,&filter->og);
			}
		}
	}
	return lBytes;
}

__declspec(dllexport) void FAR PASCAL CloseFilterOutput(HANDLE hOutput)
{
	FILTER_OUTPUT *filter=(FILTER_OUTPUT*)hOutput;
	if(filter)
	{
		int result=0;
		vorbis_analysis_wrote(&filter->vd,0);
		while(vorbis_analysis_blockout(&filter->vd,&filter->vb)==1)
		{
			vorbis_analysis(&filter->vb,0);
			vorbis_bitrate_addblock(&filter->vb);
			while(vorbis_bitrate_flushpacket(&filter->vd,&filter->op))
			{
				ogg_stream_packetin(&filter->os,&filter->op);
				result=ogg_stream_pageout(&filter->os,&filter->og);
				while(result)
				{
					fwrite(filter->og.header,1,filter->og.header_len,filter->hFile);
					fwrite(filter->og.body,1,filter->og.body_len,filter->hFile);
					if(ogg_page_eos(&filter->og))
						break;
					result=ogg_stream_pageout(&filter->os,&filter->og);
				}
			}
		}
		FILE* hFile=filter->hFile;
		if(hFile)
			fclose(hFile);
		ogg_stream_clear(&filter->os);
		vorbis_block_clear(&filter->vb);
		vorbis_dsp_clear(&filter->vd);
		vorbis_comment_clear(&filter->vc);
		vorbis_info_clear(&filter->vi);
		delete filter;
	}
}

//////////////////////////////////////////////////////////////////////
// OPTIONAL COOL EDIT FILTER FUNCTIONS
//////////////////////////////////////////////////////////////////////

__declspec(dllexport) DWORD FAR PASCAL FilterGetOptions(HWND hWnd,
														HINSTANCE hInst,
														long lSamprate,
														WORD wChannels,
														WORD wBPS,
														DWORD dwOptions)
{
	FARPROC lpfnDIALOGMsgProc=GetProcAddress(hInst,(LPCSTR)"DIALOGMsgProc");
	DWORD nDialogReturn=(DWORD)DialogBoxParam((HINSTANCE)hInst,(LPCSTR)MAKEINTRESOURCE(IDD_OPTIONS),hWnd,(DLGPROC)lpfnDIALOGMsgProc,dwOptions);
	if(nDialogReturn>0)
		return nDialogReturn;
	else
		return dwOptions;
}

__declspec(dllexport) DWORD FAR PASCAL FilterOptions(HANDLE hInput)
{
	FILTER_INPUT *filter=(FILTER_INPUT*)hInput;
	if(filter)
		return filter->options;
	else
		return 0;
}

__declspec(dllexport) DWORD FAR PASCAL FilterOptionsString(HANDLE hInput, LPSTR lpszString)
{
	FILTER_INPUT *filter=(FILTER_INPUT*)hInput;
	if(!filter)
		return 0;
	vorbis_info* vi=ov_info(&filter->vfile,-1);
	char str[128]="Ogg Vorbis ";
	char buffer[16]="";
	int version=0;
	vorbis_comment* comment=ov_comment(&filter->vfile,-1);
	if(comment&&comment->vendor)
	{
		char* v=strstr(comment->vendor,"20");
		if(v)
			version=atoi(v);
	}
	switch(version)
	{
	case 20000508:
		strcat(str,"Beta 2");
		break;
	case 20001031:
		strcat(str,"Beta 3");
		break;
	case 20010225:
		strcat(str,"Beta 4");
		break;
	case 20010615:
		strcat(str,"RC1");
		break;
	case 20010813:
		strcat(str,"RC2");
		break;
	case 20011217:
	case 20011231:
		strcat(str,"RC3");
		break;
	case 20020717:
	case 20030308:
		strcat(str,"1.0");
		break;
	case 20030909:
	case 20031230:
		strcat(str,"1.01");
		break;
	case 20040629:
	case 20040920:
		strcat(str,"1.1.0");
		break;
	case 20050304:
		strcat(str,"1.1.2");
		break;
	case 20040420:
		strcat(str,"aoTuV b2");
		break;
	case 20041120:
		strcat(str,"aoTuV b3");
		break;
	case 20050617:
		strcat(str,"aoTuV b4");
		break;
	case 20051105:
		strcat(str,"aoTuV b4.5");
		break;
	case 20051117:
		strcat(str,"aoTuV R1");
		break;
	case 20060321:
		strcat(str,"aoTuV b5");
		break;
	case 20070622:
		strcat(str,"1.2.0");
	}
	strcat(str,"\nNominal bitrate: ");
	if(vi->bitrate_nominal>=0)
	{
		itoa(vi->bitrate_nominal/1000,buffer,10);
		strcat(str,buffer);
		strcat(str," kbps");
	}
	else
		strcat(str,"unknown");
	strcat(str,"\nAverage bitrate: ");
	if(ov_bitrate(&filter->vfile,-1)>=0)
	{
		itoa(ov_bitrate(&filter->vfile,-1)/1000,buffer,10);
		strcat(str,buffer);
		strcat(str," kbps");
	}
	else
		strcat(str,"unknown");
	strcpy(lpszString,str);
	return 0;
}

__declspec(dllexport) DWORD FAR PASCAL FilterWriteSpecialData(HANDLE hOutput, LPCSTR szListType, LPCSTR szType, char * pData, DWORD dwSize)
{
	FILTER_OUTPUT *filter=(FILTER_OUTPUT*)hOutput;
	if(filter)
	{
		if((!strcmp(szListType,FCC_INFO))||(!strcmp(szListType,FCC_WAVE)&&!strcmp(szType,FCC_DISP)&&dwSize>4))
		{
			char* utf;
			if(ANSItoUTF8(strcmp(szListType,FCC_INFO)?pData+4:pData,&utf))
			{
				for(int i=0;i<TAGCOUNT;i++)
				{
					if(!strcmp(szType,TAGS[i].riff))
					{
						vorbis_comment_add_tag(&filter->vc,(char*)(TAGS[i].tag),utf);
						break;
					}
				}
				delete [] utf;
			}
		}
	}
	return 1;
}

__declspec(dllexport) DWORD FAR PASCAL FilterGetFirstSpecialData(HANDLE hInput,	SPECIALDATA * psp)
{
	FILTER_INPUT *filter=(FILTER_INPUT*)hInput;
	if(filter)
	{
		psp->hSpecialData=0;
		vorbis_comment* comment=ov_comment(&filter->vfile,-1);
		if(GetNextComment(comment,psp))
			return 1;
	}
	return 0;
}

__declspec(dllexport) DWORD FAR PASCAL FilterGetNextSpecialData(HANDLE hInput, SPECIALDATA * psp)
{
	FILTER_INPUT *filter=(FILTER_INPUT*)hInput;
	if(filter)
	{
		vorbis_comment* comment=ov_comment(&filter->vfile,-1);
		if(GetNextComment(comment,psp))
			return 1;
	}
	return 0;
}

__declspec(dllexport) BOOL FAR PASCAL DIALOGMsgProc(HWND hWndDlg, UINT Message, WPARAM wParam, LPARAM lParam)
{
	DWORD options;
	LPNMHDR pnmh;
	switch(Message)
	{
	case WM_INITDIALOG:
 		SendDlgItemMessage(hWndDlg,IDC_QUALSLIDER,TBM_SETRANGEMIN,0,0);
		SendDlgItemMessage(hWndDlg,IDC_QUALSLIDER,TBM_SETRANGEMAX,0,MAX_VBR);
		SendDlgItemMessage(hWndDlg,IDC_QUALSLIDER,TBM_SETTICFREQ,100,0);
		SendDlgItemMessage(hWndDlg,IDC_QUALSLIDER,TBM_SETPAGESIZE,0,100);
		SendDlgItemMessage(hWndDlg,IDC_NOMINALSPIN,UDM_SETRANGE32,0,MAX_ABR);
		SendDlgItemMessage(hWndDlg,IDC_MINIMUMSPIN,UDM_SETRANGE32,0,MAX_ABR);
		SendDlgItemMessage(hWndDlg,IDC_MAXIMUMSPIN,UDM_SETRANGE32,0,MAX_ABR);
		options=(DWORD)lParam;
		if((options>>30)&1)
		{
			EnableOptions(hWndDlg,0);
			SendDlgItemMessage(hWndDlg,IDC_QUALSLIDER,TBM_SETPOS,TRUE,DEFAULT_VBR);
			SendDlgItemMessage(hWndDlg,IDC_NOMINALSPIN,UDM_SETPOS,0,options&0x1ff);
			if((options>>29)&1)
			{
				if(((options>>9)&0x1ff)>(options&0x1ff))
					SendDlgItemMessage(hWndDlg,IDC_MINIMUMSPIN,UDM_SETPOS,0,options&0x1ff);
				else
					SendDlgItemMessage(hWndDlg,IDC_MINIMUMSPIN,UDM_SETPOS,0,(options>>9)&0x1ff);
				SendDlgItemMessage(hWndDlg,IDC_MINIMUMCHECK,BM_SETCHECK,BST_CHECKED,0);
			}
			else
				SendDlgItemMessage(hWndDlg,IDC_MINIMUMSPIN,UDM_SETPOS,0,options&0x1ff);
			if((options>>28)&1)
			{
				if(((options>>18)&0x1ff)<(options&0x1ff))
					SendDlgItemMessage(hWndDlg,IDC_MAXIMUMSPIN,UDM_SETPOS,0,options&0x1ff);
				else
					SendDlgItemMessage(hWndDlg,IDC_MAXIMUMSPIN,UDM_SETPOS,0,(options>>18)&0x1ff);
				SendDlgItemMessage(hWndDlg,IDC_MAXIMUMCHECK,BM_SETCHECK,BST_CHECKED,0);
			}
			else
				SendDlgItemMessage(hWndDlg,IDC_MAXIMUMSPIN,UDM_SETPOS,0,options&0x1ff);
			SendDlgItemMessage(hWndDlg,IDC_USEMBR,BM_SETCHECK,BST_CHECKED,0);
		}
		else							
		{
			EnableOptions(hWndDlg,1);
			SendDlgItemMessage(hWndDlg,IDC_NOMINALSPIN,UDM_SETPOS,0,DEFAULT_ABR);
			SendDlgItemMessage(hWndDlg,IDC_MINIMUMSPIN,UDM_SETPOS,0,DEFAULT_ABR);
			SendDlgItemMessage(hWndDlg,IDC_MAXIMUMSPIN,UDM_SETPOS,0,DEFAULT_ABR);
			if((!options)||((options&0xffff)>MAX_VBR))
				SendDlgItemMessage(hWndDlg,IDC_QUALSLIDER,TBM_SETPOS,TRUE,DEFAULT_VBR);
			else
				SendDlgItemMessage(hWndDlg,IDC_QUALSLIDER,TBM_SETPOS,TRUE,options&0xffff);
			SendDlgItemMessage(hWndDlg,IDC_USEVBR,BM_SETCHECK,BST_CHECKED,0);
		}
		return TRUE;
    case WM_CLOSE:
		PostMessage(hWndDlg, WM_COMMAND, IDCANCEL, 0L);
		return TRUE;
	case WM_NOTIFY:
		pnmh=(LPNMHDR)lParam;
		if(pnmh->idFrom==IDC_QUALSLIDER)
		{
			float f=(float)SendDlgItemMessage(hWndDlg,IDC_QUALSLIDER,TBM_GETPOS,0,0)-100;
			f/=100;
			char buffer[32]="";
			sprintf(buffer,"Quality: %.2f",f);
			SetDlgItemText(hWndDlg,IDC_QUALITY,buffer);
		}
		break;
    case WM_COMMAND:
		options=0;
		switch(LOWORD(wParam))
        {
			DWORD dw;
			case IDOK:
				options=SendDlgItemMessage(hWndDlg,IDC_USEVBR,BM_GETCHECK,0,0);
				if(options==BST_CHECKED)
				{
					options=0x80000000;
					options+=SendDlgItemMessage(hWndDlg,IDC_QUALSLIDER,TBM_GETPOS,0,0);
				}
				else
				{
					dw=SendDlgItemMessage(hWndDlg,IDC_MAXIMUMCHECK,BM_GETCHECK,0,0);
					if(dw==BST_CHECKED)
					{
						dw=SendDlgItemMessage(hWndDlg,IDC_MAXIMUMSPIN,UDM_GETPOS,0,0);
						if(dw>MAX_ABR)
							dw=MAX_ABR;
					}
					options=dw;
					options<<=9;
					dw=SendDlgItemMessage(hWndDlg,IDC_MINIMUMCHECK,BM_GETCHECK,0,0);
					if(dw==BST_CHECKED)
					{
						dw=SendDlgItemMessage(hWndDlg,IDC_MINIMUMSPIN,UDM_GETPOS,0,0);
						if(dw>MAX_ABR)
							dw=MAX_ABR;
					}
					options+=dw;
					options<<=9;
					dw=SendDlgItemMessage(hWndDlg,IDC_NOMINALSPIN,UDM_GETPOS,0,0);
					if(dw>MAX_ABR)
						dw=MAX_ABR;
					options+=dw;
					dw=SendDlgItemMessage(hWndDlg,IDC_MAXIMUMCHECK,BM_GETCHECK,0,0);
					if(dw==BST_CHECKED)
						options|=0x10000000;
					dw=SendDlgItemMessage(hWndDlg,IDC_MINIMUMCHECK,BM_GETCHECK,0,0);
					if(dw==BST_CHECKED)
						options|=0x20000000;
					options|=0x40000000;
				}
				EndDialog(hWndDlg,options);
				return TRUE;
            case IDCANCEL:
				EndDialog(hWndDlg,0);
				return TRUE;
			case IDC_USEVBR:
				if(SendDlgItemMessage(hWndDlg,IDC_USEVBR,BM_GETCHECK,0,0)==BST_CHECKED)
					EnableOptions(hWndDlg,1);
				else
					EnableOptions(hWndDlg,0);
				break;
			case IDC_USEMBR:
				if(SendDlgItemMessage(hWndDlg,IDC_USEMBR,BM_GETCHECK,0,0)==BST_CHECKED)
					EnableOptions(hWndDlg,0);
				else
					EnableOptions(hWndDlg,1);
		}
	}
	return FALSE;
}

//////////////////////////////////////////////////////////////////////
// INTERNAL FUNCTIONS
//////////////////////////////////////////////////////////////////////

void EnableOptions(HWND hWndDlg, BOOL vbr)
{
	if(vbr)
	{
		EnableWindow(GetDlgItem(hWndDlg,IDC_NOMINALTEXT),FALSE);
		EnableWindow(GetDlgItem(hWndDlg,IDC_NOMINAL),FALSE);
		EnableWindow(GetDlgItem(hWndDlg,IDC_NOMINALSPIN),FALSE);
		EnableWindow(GetDlgItem(hWndDlg,IDC_MINIMUMCHECK),FALSE);
		EnableWindow(GetDlgItem(hWndDlg,IDC_MINIMUM),FALSE);
		EnableWindow(GetDlgItem(hWndDlg,IDC_MINIMUMSPIN),FALSE);
		EnableWindow(GetDlgItem(hWndDlg,IDC_MAXIMUMCHECK),FALSE);
		EnableWindow(GetDlgItem(hWndDlg,IDC_MAXIMUM),FALSE);
		EnableWindow(GetDlgItem(hWndDlg,IDC_MAXIMUMSPIN),FALSE);
		EnableWindow(GetDlgItem(hWndDlg,IDC_QUALSLIDER),TRUE);
		EnableWindow(GetDlgItem(hWndDlg,IDC_QUALITY),TRUE);
	}
	else
	{
		EnableWindow(GetDlgItem(hWndDlg,IDC_QUALSLIDER),FALSE);
		EnableWindow(GetDlgItem(hWndDlg,IDC_QUALITY),FALSE);
		EnableWindow(GetDlgItem(hWndDlg,IDC_NOMINALTEXT),TRUE);
		EnableWindow(GetDlgItem(hWndDlg,IDC_NOMINAL),TRUE);
		EnableWindow(GetDlgItem(hWndDlg,IDC_NOMINALSPIN),TRUE);
		EnableWindow(GetDlgItem(hWndDlg,IDC_MINIMUMCHECK),TRUE);
		EnableWindow(GetDlgItem(hWndDlg,IDC_MINIMUM),TRUE);
		EnableWindow(GetDlgItem(hWndDlg,IDC_MINIMUMSPIN),TRUE);
		EnableWindow(GetDlgItem(hWndDlg,IDC_MAXIMUMCHECK),TRUE);
		EnableWindow(GetDlgItem(hWndDlg,IDC_MAXIMUM),TRUE);
		EnableWindow(GetDlgItem(hWndDlg,IDC_MAXIMUMSPIN),TRUE);
	}
}

BOOL GetNextComment(vorbis_comment* comment, SPECIALDATA* psp)
{
	for(int i=(int)(void*)psp->hSpecialData;i<comment->comments;i++)
	{
		psp->hSpecialData=(void*)(i+1);
		char* e=strchr(comment->user_comments[i],'=');
		if(e&&e-comment->user_comments[i]&&strlen(e+1))
		{
			char* t=new char[e-comment->user_comments[i]+1];
			if(t)
			{
				ZeroMemory(t,e-comment->user_comments[i]+1);
				strncpy(t,comment->user_comments[i],e-comment->user_comments[i]);
				for(int j=0;j<TAGCOUNT;j++)
				{
					if(!_stricmp(t,TAGS[j].tag))
					{
						char* ansi;
						if(UTF8toANSI(e+1,&ansi))
						{
							int x=strlen(ansi)+2;
							BOOL bDisp=0;
							if(!strcmp(TAGS[j].riff,FCC_DISP))
							{
								bDisp=1;
								x+=4;
							}
							psp->hData=GlobalAlloc(GMEM_MOVEABLE|GMEM_ZEROINIT,x);
							if(psp->hData)
							{
								char* pData=(char*)GlobalLock(psp->hData);
								psp->dwSize=x-1;
								psp->dwExtra=1;
								if(bDisp)
								{
									strcpy(psp->szListType,FCC_WAVE);
									strcpy(pData+4,ansi);
									pData[0]=1;
								}
								else
								{
									strcpy(psp->szListType,FCC_INFO);
									strcpy(pData,ansi);
								}
								strcpy(psp->szType,TAGS[j].riff);
								GlobalUnlock(pData);
								delete [] ansi;
								delete [] t;
								return 1;
							}
							delete [] ansi;
						}
						break;
					}
				}
				delete [] t;
			}
		}
	}
	return 0;
}

BOOL ANSItoUTF8(char* ansi, char** utf8)
{
	BOOL success=0;
	int i=MultiByteToWideChar(CP_ACP,MB_PRECOMPOSED,ansi,-1,0,0);
	if(i>1)
	{
		WCHAR* wBuf=new WCHAR[i];
		if(wBuf)
		{
			MultiByteToWideChar(CP_ACP,MB_PRECOMPOSED,ansi,-1,wBuf,i);
			i=WideCharToMultiByte(CP_UTF8,0,wBuf,-1,0,0,0,0);
			if(i>1)
			{
				char* buf=new char[i];
				if(buf)
				{
					WideCharToMultiByte(CP_UTF8,0,wBuf,-1,buf,i,0,0);
					*utf8=buf;
					success=1;
				}
			}
			delete [] wBuf;
		}
	}
	return success;
}

BOOL UTF8toANSI(char* utf8, char** ansi)
{
	BOOL success=0;
	int i=MultiByteToWideChar(CP_UTF8,0,utf8,-1,0,0);
	if(i>1)
	{
		WCHAR* wBuf=new WCHAR[i];
		if(wBuf)
		{
			MultiByteToWideChar(CP_UTF8,0,utf8,-1,wBuf,i);
			EnsureCRLF(&wBuf);
			i=WideCharToMultiByte(CP_ACP,0,wBuf,-1,0,0,0,0);
			if(i>1)
			{
				char* buf=new char[i];
				if(buf)
				{
					WideCharToMultiByte(CP_ACP,0,wBuf,-1,buf,i,0,0);
					*ansi=buf;
					success=1;
				}
			}
			delete [] wBuf;
		}
	}
	return success;
}

void EnsureCRLF(WCHAR** str)
{
	WCHAR* a=*str;
	WCHAR* b;
	int i;
	while(b=wcsstr(a,L"\r\n"))
	{
		a=++b+1;
		i=wcslen(a);
		if(i)
			memmove(b,a,i*2);
		*(b+i)=0;
	}
	a=*str;
	i=0;
	while(b=wcspbrk(a,L"\r\n"))
	{
		a=b+1;
		i++;
	}
	if(i)
	{
		a=*str;
		i=wcslen(a)+i+1;
		WCHAR* w=new WCHAR[i];
		if(w)
		{
			ZeroMemory(w,i*sizeof(WCHAR));
			while(b=wcspbrk(a,L"\r\n"))
			{
				wcsncat(w,a,b-a);
				wcscat(w,L"\r\n");
				a=b+1;
			}
			wcscat(w,a);
			delete [] *str;
			*str=w;
		}
	}
}
