/*
QuickTime File Filter for Cool Edit/Adobe Audition
Copyright  2006 James Chapman
http://www.vuplayer.com

QuickTime is Copyright  2006 Apple Computer, Inc
http://www.apple.com/quicktime
*/

#include <windows.h>
#include <stdio.h>
#include <qtml.h>
#include <movies.h>
#include "../filters.h"

#define CHUNKSIZE			65536

typedef struct	// any special vars associated with input file
{
	DWORD flags;
	DWORD samplecount;
	DWORD samplerate;
	WORD channels;
	WORD bps;
	long moviesize;
	short movieref;
	Movie movie;
	MovieAudioExtractionRef session;
	BYTE buffer[CHUNKSIZE];
} FILTER_INPUT;

BOOL QT_Init()	// Loading & unloading the QuickTime libraries is slow, but this can't be done in DLLMain
{
	if(InitializeQTML(0)==noErr)
	{
		if(EnterMovies()==noErr)
			return 1;
		TerminateQTML();
	}
	return 0;
}

void QT_Free()
{
	ExitMovies();
	TerminateQTML();
}

BOOL QT_OpenFile(LPSTR filename, FILTER_INPUT* filter)
{
	FSSpec fss;
	if(NativePathNameToFSSpec(filename,&fss,kErrorIfFileNotFound)==noErr)
	{
		if(OpenMovieFile(&fss,&filter->movieref,fsRdPerm)==noErr)
		{
			short resid=0;
			if(NewMovieFromFile(&filter->movie,filter->movieref,&resid,0,newMovieActive,0)==noErr)
			{
				if(MovieAudioExtractionBegin(filter->movie,0,&filter->session)==noErr)
				{
					AudioStreamBasicDescription asbd;
					if(MovieAudioExtractionGetProperty(
						filter->session,
						kQTPropertyClass_MovieAudioExtraction_Audio,
						kQTMovieAudioExtractionAudioPropertyID_AudioStreamBasicDescription,
						sizeof(AudioStreamBasicDescription),&asbd,0)==noErr)
					{
						filter->samplerate=(DWORD)asbd.mSampleRate;
						filter->bps=(WORD)asbd.mBitsPerChannel;
						filter->channels=(WORD)asbd.mChannelsPerFrame;
						TimeValue t=GetMovieDuration(filter->movie);
						TimeScale ts=GetMovieTimeScale(filter->movie);
						if(ts)
							filter->samplecount=(DWORD)((float)t*filter->samplerate/ts+0.5f);
						filter->moviesize=GetMovieDataSize(filter->movie,0,t);
						return 1;	// Use the default non-interleaved floating point format ASBD
					}
					MovieAudioExtractionEnd(filter->session);
				}
				DisposeMovie(filter->movie);
			}
			CloseMovieFile(filter->movieref);
		}
	}
	return 0;
}

void QT_CloseFile(FILTER_INPUT* filter)
{
	MovieAudioExtractionEnd(filter->session);
	DisposeMovie(filter->movie);
	CloseMovieFile(filter->movieref);
}

//////////////////////////////////////////////////////////////////////
// REQUIRED COOL EDIT FILTER FUNCTIONS
//////////////////////////////////////////////////////////////////////

__declspec(dllexport) short FAR PASCAL QueryCoolFilter(COOLQUERY* cq)
{
	if(QT_Init())
		QT_Free();
	else
		return 0;
	ZeroMemory(cq,sizeof(COOLQUERY));
	strcpy(cq->szName,"Apple QuickTime");		
	strcpy(cq->szCopyright,"QuickTime  Apple Computer, Inc");
	strcpy(cq->szExt,"m4a");
	strcpy(cq->szExt2,"mp4");
	cq->lChunkSize=CHUNKSIZE;
	cq->dwFlags=QF_CANLOAD|QF_CANDO32BITFLOATS;
	cq->Stereo16=0xFF;
	cq->Stereo32=0xFF;
 	return C_VALIDLIBRARY;
}

__declspec(dllexport) BOOL FAR PASCAL FilterUnderstandsFormat(LPSTR filename)
{
	BOOL success=0;
	if(QT_Init())
	{
		FILTER_INPUT* filter=new FILTER_INPUT;
		if(filter)
		{
			ZeroMemory(filter,sizeof(FILTER_INPUT));
			if(QT_OpenFile(filename,filter))
			{
				QT_CloseFile(filter);
				success=1;
			}
			delete filter;
		}
		QT_Free();
	}
	return success;
}

//////////////////////////////////////////////////////////////////////
// REQUIRED COOL EDIT FILTER FUNCTIONS - READING
//////////////////////////////////////////////////////////////////////

__declspec(dllexport) DWORD FAR PASCAL FilterGetFileSize(HANDLE hInput)
{
	FILTER_INPUT* filter=(FILTER_INPUT*)hInput;
	return filter->samplecount*filter->channels*(filter->bps/8);
}

__declspec(dllexport) HANDLE FAR PASCAL OpenFilterInput(LPSTR filename,long* lRate,WORD* wBPS,WORD* wChannels,HWND hWnd,long* lChunkSize)
{
	if(QT_Init())
	{
		FILTER_INPUT *filter=new FILTER_INPUT;
		if(filter)
		{
			ZeroMemory(filter,sizeof(FILTER_INPUT));
			if(QT_OpenFile(filename,filter))
			{
				*lRate=filter->samplerate;
				*wBPS=filter->bps;
				*wChannels=filter->channels;
				*lChunkSize=CHUNKSIZE;
				return filter;
			}
			delete filter;
		}
		QT_Free();
	}
	return 0;
}


__declspec(dllexport) DWORD FAR PASCAL ReadFilterInput(HANDLE hInput, unsigned char* buf, long lBytes)
{
	UInt32 numframes=0;
	FILTER_INPUT *filter=(FILTER_INPUT*)hInput;
	AudioBufferList* buflist=new AudioBufferList[filter->channels];
	if(buflist)
	{
		buflist->mNumberBuffers=filter->channels;
		for(UINT i=0;i<buflist->mNumberBuffers;i++)
		{
			buflist->mBuffers[i].mNumberChannels=1;
			buflist->mBuffers[i].mDataByteSize=lBytes/buflist->mNumberBuffers;
			buflist->mBuffers[i].mData=filter->buffer+lBytes/buflist->mNumberBuffers*i;
		}
		numframes=lBytes/filter->channels/(filter->bps/8);
		if(MovieAudioExtractionFillBuffer(filter->session,&numframes,buflist,&filter->flags)==noErr)
		{
			float* outbuf=(float*)buf;
			float* inbuf=(float*)filter->buffer;
			for(UINT a=0;a<numframes;a++)				// Interleave & normalize samples
			{
				for(UINT b=0;b<filter->channels;b++)
					outbuf[a*filter->channels+b]=inbuf[b*numframes+a]*32768.0f;
			}
			numframes*=filter->channels*(filter->bps/8);
		}
		else
			numframes=0;
		delete [] buflist;
	}
	return numframes;
}

__declspec(dllexport) void FAR PASCAL CloseFilterInput(HANDLE hInput)
{
	FILTER_INPUT *filter=(FILTER_INPUT*)hInput;
	if(filter)
	{
		QT_CloseFile(filter);
		QT_Free();
		delete filter;
	}
}

//////////////////////////////////////////////////////////////////////
// OPTIONAL COOL EDIT FILTER FUNCTIONS
//////////////////////////////////////////////////////////////////////

__declspec(dllexport) DWORD FAR PASCAL FilterOptionsString(HANDLE hInput, LPSTR lpszString)
{
	FILTER_INPUT* filter=(FILTER_INPUT*)hInput;
	if(filter)
	{
		Track t=GetMovieIndTrackType(filter->movie,1,SoundMediaType,movieTrackMediaType);
		if(t)
		{
			Media m=GetTrackMedia(t);
			if(m)
			{
				SampleDescriptionHandle h=(SampleDescriptionHandle)NewHandle(0);
				if(h)
				{
					GetMediaSampleDescription(m,1,h);
					switch((*h)->dataFormat)
					{
					case(FOUR_CHAR_CODE('mp4a')):
						strcpy(lpszString,"MPEG-4 AAC\n");
						break;
					case(FOUR_CHAR_CODE('alac')):
						strcpy(lpszString,"Apple Lossless\n");
						break;
					default:
						strcpy(lpszString,"Apple QuickTime\n");
					}
					DisposeHandle((Handle)h);
				}
				if(filter->samplecount)
				{
					long bitrate=(long)((float)filter->moviesize/filter->samplecount*filter->samplerate*8/1000+0.5f);
					if(bitrate)
					{
						char buf[32];
						sprintf(buf,"Bitrate: %d kbps",bitrate);
						strcat(lpszString,buf);
					}
				}
				DisposeTrackMedia(m);
			}
		}
	}
	return 0;
}
