/*
Windows Media Audio File Filter for Cool Edit/Adobe Audition
Copyright  2006 James Chapman
http://www.vuplayer.com

Windows Media Audio is Copyright  1999-2006 Microsoft Corporation
http://www.microsoft.com/windowsmedia
*/

#include <windows.h>
#include <mmsystem.h>
#include <wmsdk/wmsdk.h>
#include "../filters.h"
#include "resource.h"

#define CHUNKSIZE	65536
#define FLAG_VBR	0x80000000

#define TAGCOUNT	8

#define FCC_INFO	"INFO"
#define FCC_WAVE	"WAVE"
#define FCC_DISP	"DISP"

typedef struct
{
	LPCSTR riff;
	LPCWSTR tag;
} TAGDATA;

const TAGDATA TAGS[TAGCOUNT]=
{
	{"IART",g_wszWMAuthor},
	{"ICMT",g_wszWMDescription},
	{"ICOP",g_wszWMCopyright},
	{"ICRD",g_wszWMYear},
	{"IGNR",g_wszWMGenre},
	{"INAM",g_wszWMAlbumTitle},
	{"ISFT",g_wszWMToolName},
	{"DISP",g_wszWMTitle}
};

typedef struct
{
	LPSTR filename;
	BOOL beginwriting;
	IWMWriter* pWriter;
	QWORD cnsCurrent;
	long samplerate;
	WORD bps;
	WORD channels;
	DWORD dwOptions;
} FILTER_OUTPUT;

typedef struct
{
	IWMSyncReader* pReader;
	INSSBuffer* pBuffer;
	BYTE* pData;
	LPSTR filename;
	long samplerate;
	WORD bps;
	WORD channels;
	DWORD dwPos;
	DWORD dwSize;
	DWORD dwOptions;
} FILTER_INPUT;

BOOL OpenWMAInput(FILTER_INPUT* pFilter);
long ReadWMAInput(FILTER_INPUT* pFilter, BYTE* buf, long lBytes);
void CloseWMAInput(FILTER_INPUT* pFilter);
BOOL OpenWMAOutput(FILTER_OUTPUT* pFilter);
long WriteWMAOutput(FILTER_OUTPUT* pFilter, BYTE* pBuf, long lBytes);
void CloseWMAOutput(FILTER_OUTPUT* pFilter);
BOOL GetNextComment(FILTER_INPUT* pFilter, SPECIALDATA* psp);
BOOL SetComment(FILTER_OUTPUT* pFilter, WCHAR* field, WCHAR* value);
DWORD GetDefaultSetting();
void UpdateFormatCombo(HWND hWndDlg, BOOL vbr, DWORD codecindex);
void UpdateCodecCombo(HWND hWndDlg, BOOL vbr, DWORD codecindex);
BOOL ANSItoUNICODE(char* ansi, WCHAR** unicode);
BOOL UNICODEtoANSI(WCHAR* unicode, char** ansi);
void EnsureCRLF(WCHAR** str);

//////////////////////////////////////////////////////////////////////
// REQUIRED COOL EDIT FILTER FUNCTIONS
//////////////////////////////////////////////////////////////////////

__declspec(dllexport) short FAR PASCAL QueryCoolFilter(COOLQUERY* cq)
{
	ZeroMemory(cq,sizeof(COOLQUERY));
	strcpy(cq->szName,"Windows Media Audio");		
	strcpy(cq->szCopyright,"Windows Media Audio  Microsoft Corporation");
	strcpy(cq->szExt,"wma");
	cq->lChunkSize=CHUNKSIZE; 
	cq->dwFlags=QF_CANLOAD|QF_CANSAVE|QF_RATEADJUSTABLE|QF_HASOPTIONSBOX|QF_CANDO32BITFLOATS|QF_READSPECIALFIRST|QF_WRITESPECIALFIRST;;
 	cq->Stereo8=0xff;
 	cq->Stereo16=0xff;
 	cq->Stereo32=0xff;
 	cq->Mono8=0xff;
 	cq->Mono16=0xff;
 	cq->Mono32=0xff;
 	return C_VALIDLIBRARY;
}

__declspec(dllexport) BOOL FAR PASCAL FilterUnderstandsFormat(LPSTR filename)
{
	FILTER_INPUT filter;
	ZeroMemory(&filter,sizeof(FILTER_INPUT));
	filter.filename=filename;
	BOOL success=OpenWMAInput(&filter);
	if(success)
		CloseWMAInput(&filter);
	return success;
}

//////////////////////////////////////////////////////////////////////
// REQUIRED COOL EDIT FILTER FUNCTIONS - READING
//////////////////////////////////////////////////////////////////////

__declspec(dllexport) HANDLE FAR PASCAL OpenFilterInput(LPSTR filename,long* lRate,WORD* wBPS,WORD* wChannels,HWND hWnd,long* lChunkSize)
{
	FILTER_INPUT *filter=new FILTER_INPUT;
	if(filter)
	{
		ZeroMemory(filter,sizeof(FILTER_INPUT));
		filter->filename=filename;
		BOOL success=OpenWMAInput(filter);
		if(success)
		{
			*lRate=filter->samplerate;
			*wChannels=filter->channels;
			*wBPS=filter->bps;
			*lChunkSize=CHUNKSIZE;
		}
		else
		{
			delete filter;
			filter=0;
		}
	}
	return filter;
}

__declspec(dllexport) DWORD FAR PASCAL FilterGetFileSize(HANDLE hInput)
{
	FILTER_INPUT *filter=(FILTER_INPUT*)hInput;
	QWORD duration=0;
	IWMHeaderInfo3* pHeader;
	HRESULT hr=filter->pReader->QueryInterface(IID_IWMHeaderInfo3,(void**)&pHeader);
	if(SUCCEEDED(hr))
	{
		WORD wStreamNum=0;
		WORD cbLength=0;
		WMT_ATTR_DATATYPE wmType;
		BYTE* pValue=0;
		hr=pHeader->GetAttributeByName(&wStreamNum,g_wszWMDuration,&wmType,0,&cbLength);
		if(SUCCEEDED(hr)&&cbLength)
		{
			pValue=new BYTE[cbLength];
			if(pValue)
			{
				hr=pHeader->GetAttributeByName(&wStreamNum,g_wszWMDuration,&wmType,pValue,&cbLength);
				if(SUCCEEDED(hr))							// Returned value is sometimes on the low side
					duration=*((QWORD*)pValue)/1000000+1;	// so allow a bit of leeway
				delete [] pValue;
			}
		}
		pHeader->Release();
	}
	return (DWORD)(duration*filter->samplerate*filter->channels*(filter->bps/8)/10);
}

__declspec(dllexport) DWORD FAR PASCAL ReadFilterInput(HANDLE hInput, unsigned char* buf, long lBytes)
{
	FILTER_INPUT *filter=(FILTER_INPUT*)hInput;
	return ReadWMAInput(filter,buf,lBytes);
}

__declspec(dllexport) void FAR PASCAL CloseFilterInput(HANDLE hInput)
{
	FILTER_INPUT *filter=(FILTER_INPUT*)hInput;
	if(filter)
	{
		CloseWMAInput(filter);
		delete filter;
	}
}

//////////////////////////////////////////////////////////////////////
// REQUIRED COOL EDIT FILTER FUNCTIONS - WRITING
//////////////////////////////////////////////////////////////////////

__declspec(dllexport) void FAR PASCAL GetSuggestedSampleType(long* lRate, WORD* wBPS, WORD *wChannels)
{
	*wBPS=0;
	*lRate=0;
	*wChannels=0;
}

__declspec(dllexport) HANDLE FAR PASCAL OpenFilterOutput(LPSTR filename,
														 long samprate,
														 WORD BPS,
														 WORD channels,
														 long lSize,
														 long* lChunkSize,
														 DWORD options)
{
	FILTER_OUTPUT *filter=new FILTER_OUTPUT;
	if(filter)
	{
		ZeroMemory(filter,sizeof(FILTER_OUTPUT));
		filter->filename=filename;
		filter->samplerate=samprate;
		filter->bps=(BPS==32)?24:BPS;
		filter->channels=channels;
		filter->dwOptions=options;
		if(OpenWMAOutput(filter))
			*lChunkSize=CHUNKSIZE;
		else
		{
			delete filter;
			filter=0;
		}
	}
	return filter;
}

__declspec(dllexport) DWORD FAR PASCAL WriteFilterOutput(HANDLE hOutput, BYTE* pBuf, long lBytes)
{
	FILTER_OUTPUT *filter=(FILTER_OUTPUT*)hOutput;
	return WriteWMAOutput(filter,pBuf,lBytes);
}

__declspec(dllexport) void FAR PASCAL CloseFilterOutput(HANDLE hOutput)
{
	FILTER_OUTPUT *filter=(FILTER_OUTPUT*)hOutput;
	if(filter)
	{
		CloseWMAOutput(filter);
		delete filter;
	}
}

//////////////////////////////////////////////////////////////////////
// OPTIONAL COOL EDIT FILTER FUNCTIONS
//////////////////////////////////////////////////////////////////////

__declspec(dllexport) DWORD FAR PASCAL FilterGetOptions(HWND hWnd,
														HINSTANCE hInst,
														long lSamprate,
														WORD wChannels,
														WORD wBPS,
														DWORD dwOptions)
{
	FARPROC lpfnDIALOGMsgProc=GetProcAddress(hInst,(LPCSTR)"DIALOGMsgProc");
	DWORD nDialogReturn=(DWORD)DialogBoxParam((HINSTANCE)hInst,(LPCSTR)MAKEINTRESOURCE(IDD_OPTIONS),hWnd,(DLGPROC)lpfnDIALOGMsgProc,dwOptions);
	if(nDialogReturn>0)
		return nDialogReturn;
	else
		return dwOptions;
}

__declspec(dllexport) DWORD FAR PASCAL FilterOptions(HANDLE hInput)
{
	FILTER_INPUT *filter=(FILTER_INPUT*)hInput;
	if(filter)
		return filter->dwOptions;
	else
		return 0;
}

__declspec(dllexport) DWORD FAR PASCAL FilterOptionsString(HANDLE hInput, LPSTR lpszString)
{
	FILTER_INPUT *filter=(FILTER_INPUT*)hInput;
	if(filter)
	{
		IWMHeaderInfo3* pHeader;
		HRESULT hr=filter->pReader->QueryInterface(IID_IWMHeaderInfo3,(void**)&pHeader);
		if(SUCCEEDED(hr))
		{
			WORD pcchName=0;
			WORD pcchDesc=0;
			WORD pcbCodecInfo=0;
			WMT_CODEC_INFO_TYPE pCodecType;
			hr=pHeader->GetCodecInfo(0,&pcchName,0,&pcchDesc,0,&pCodecType,&pcbCodecInfo,0);
			if(SUCCEEDED(hr))
			{
				WCHAR* pwszName=new WCHAR[pcchName];
				WCHAR* pwszDesc=new WCHAR[pcchDesc];
				BYTE* pbCodecInfo=new BYTE[pcbCodecInfo];
				if(pwszName&&pwszDesc&&pbCodecInfo)
				{
					hr=pHeader->GetCodecInfo(0,&pcchName,pwszName,&pcchDesc,pwszDesc,&pCodecType,&pcbCodecInfo,pbCodecInfo);
					if(SUCCEEDED(hr))
					{
						WideCharToMultiByte(CP_ACP,0,pwszName,-1,lpszString,64,0,0);
						strcat(lpszString,"\n");
						WCHAR* w=wcschr(pwszDesc,',');
						if(w)
							*w=0;
						WideCharToMultiByte(CP_ACP,0,pwszDesc,-1,lpszString+strlen(lpszString),128-strlen(lpszString),0,0);
					}
				}
				if(pwszName)
					delete [] pwszName;
				if(pwszDesc)
					delete [] pwszDesc;
				if(pbCodecInfo)
					delete [] pbCodecInfo;
			}
			pHeader->Release();
		}
	}
    return 0;
}

__declspec(dllexport) DWORD FAR PASCAL FilterWriteSpecialData(HANDLE hOutput, LPCSTR szListType, LPCSTR szType, char * pData, DWORD dwSize)
{
	FILTER_OUTPUT *filter=(FILTER_OUTPUT*)hOutput;
	if(filter)
	{
		if((!strcmp(szListType,FCC_INFO))||(!strcmp(szListType,FCC_WAVE)&&!strcmp(szType,FCC_DISP)&&dwSize>4))
		{
			WCHAR* unicode;
			if(ANSItoUNICODE(strcmp(szListType,FCC_INFO)?pData+4:pData,&unicode))
			{
				for(int i=0;i<TAGCOUNT;i++)
				{
					if(!strcmp(szType,TAGS[i].riff))
					{
						SetComment(filter,(WCHAR*)TAGS[i].tag,unicode);
						break;
					}
				}
				delete [] unicode;
			}
		}
	}
	return 1;
}

__declspec(dllexport) DWORD FAR PASCAL FilterGetFirstSpecialData(HANDLE hInput,	SPECIALDATA * psp)
{
	FILTER_INPUT *filter=(FILTER_INPUT*)hInput;
	if(filter)
	{
		psp->hSpecialData=0;
		if(GetNextComment(filter,psp))
			return 1;
	}
	return 0;
}

__declspec(dllexport) DWORD FAR PASCAL FilterGetNextSpecialData(HANDLE hInput, SPECIALDATA * psp)
{
	FILTER_INPUT *filter=(FILTER_INPUT*)hInput;
	if(filter)
	{
		if(GetNextComment(filter,psp))
			return 1;
	}
	return 0;
}

__declspec(dllexport) BOOL FAR PASCAL DIALOGMsgProc(HWND hWndDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch(message)
	{
	case WM_INITDIALOG:
		{
			SendDlgItemMessage(hWndDlg,IDC_VBR,CB_ADDSTRING,0,(LPARAM)"CBR");
			SendDlgItemMessage(hWndDlg,IDC_VBR,CB_ADDSTRING,0,(LPARAM)"VBR");
			DWORD options=(DWORD)lParam;
			if(!options)
				options=GetDefaultSetting();
			BOOL vbr=0;
			if(options&FLAG_VBR)
				vbr=1;
			SendDlgItemMessage(hWndDlg,IDC_VBR,CB_SETCURSEL,vbr,0);
			DWORD codecindex=(options>>16)&0xff;
			UpdateCodecCombo(hWndDlg,vbr,codecindex);
			codecindex=options&0xffff;
			DWORD itemcount=SendDlgItemMessage(hWndDlg,IDC_FORMAT,CB_GETCOUNT,0,0);
			for(DWORD index=0;index<itemcount;index++)
			{
				if(SendDlgItemMessage(hWndDlg,IDC_FORMAT,CB_GETITEMDATA,index,0)==(long)codecindex)
				{
					SendDlgItemMessage(hWndDlg,IDC_FORMAT,CB_SETCURSEL,index,0);
					break;
				}
			}
		}
		return 1;
	case WM_CLOSE:
		PostMessage(hWndDlg,WM_COMMAND,IDCANCEL,0);
		return 1;
	case WM_COMMAND:
		switch(LOWORD(wParam))
		{
			case IDOK:
				{
					DWORD config=SendDlgItemMessage(hWndDlg,IDC_CODEC,CB_GETITEMDATA,SendDlgItemMessage(hWndDlg,IDC_CODEC,CB_GETCURSEL,0,0),0);
					config<<=16;
					config+=SendDlgItemMessage(hWndDlg,IDC_FORMAT,CB_GETITEMDATA,SendDlgItemMessage(hWndDlg,IDC_FORMAT,CB_GETCURSEL,0,0),0);
					if(SendDlgItemMessage(hWndDlg,IDC_VBR,CB_GETCURSEL,0,0))
						config|=FLAG_VBR;
					EndDialog(hWndDlg,config);
				}
				return 1;
			case IDCANCEL:
				EndDialog(hWndDlg,0);
				return 1;
			case IDC_VBR:
				if(HIWORD(wParam)==CBN_SELCHANGE)
					UpdateCodecCombo(hWndDlg,SendDlgItemMessage(hWndDlg,IDC_VBR,CB_GETCURSEL,0,0),0);
				return 1;
			case IDC_CODEC:
				if(HIWORD(wParam)==CBN_SELCHANGE)
					UpdateFormatCombo(hWndDlg,SendDlgItemMessage(hWndDlg,IDC_VBR,CB_GETCURSEL,0,0),SendDlgItemMessage(hWndDlg,IDC_CODEC,CB_GETITEMDATA,SendDlgItemMessage(hWndDlg,IDC_CODEC,CB_GETCURSEL,0,0),0));
				return 1;
		}
 	}
	return 0;
}

//////////////////////////////////////////////////////////////////////
// INTERNAL FUNCTIONS
//////////////////////////////////////////////////////////////////////

BOOL OpenWMAInput(FILTER_INPUT* pFilter)
{
	HRESULT hr=CoInitializeEx(0,COINIT_APARTMENTTHREADED);
	if(FAILED(hr))
		return 0;
	BOOL success=0;
	hr=WMCreateSyncReader(0,WMT_RIGHT_PLAYBACK,&pFilter->pReader);
	if(SUCCEEDED(hr))
	{
		int wLen=MultiByteToWideChar(CP_ACP,MB_PRECOMPOSED,pFilter->filename,-1,0,0);
		WCHAR* wBuf=new WCHAR[wLen];
		if(wBuf)
		{
			MultiByteToWideChar(CP_ACP,MB_PRECOMPOSED,pFilter->filename,-1,wBuf,wLen);
			hr=pFilter->pReader->Open(wBuf);
			delete [] wBuf;
			if(SUCCEEDED(hr))
			{
				WMT_ATTR_DATATYPE datatype=WMT_TYPE_BOOL;
				DWORD i=1;
				hr=pFilter->pReader->SetOutputSetting(0,g_wszEnableDiscreteOutput,WMT_TYPE_BOOL,(LPBYTE)&i,sizeof(DWORD));
				i=2;
				hr=pFilter->pReader->SetOutputSetting(0,g_wszSpeakerConfig,WMT_TYPE_DWORD,(LPBYTE)&i,sizeof(DWORD));
				IWMOutputMediaProps* pProps;
				hr=pFilter->pReader->GetOutputFormat(0,0,&pProps);
				if(SUCCEEDED(hr))
				{
					hr=pProps->GetMediaType(0,&i);
					if(SUCCEEDED(hr))
					{
						BYTE* buffer=new BYTE[i];
						if(buffer)
						{
							WM_MEDIA_TYPE* pMediaType=(WM_MEDIA_TYPE*)buffer;
							hr=pProps->GetMediaType(pMediaType,&i);
							if(SUCCEEDED(hr))
							{
								if(pMediaType->majortype==WMMEDIATYPE_Audio)
								{
									WAVEFORMATEX* pwfx=(WAVEFORMATEX*)pMediaType->pbFormat;
									pFilter->channels=pwfx->nChannels;
									pFilter->samplerate=pwfx->nSamplesPerSec;
									pFilter->bps=(pwfx->wBitsPerSample==24)?32:pwfx->wBitsPerSample;
									if(pFilter->channels&&pFilter->samplerate&&pFilter->bps)
									{
										hr=pFilter->pReader->SetOutputProps(0,pProps);
										if(SUCCEEDED(hr))
											success=1;
									}
								}
							}
							delete [] buffer;
						}
					}
					pProps->Release();
				}
			}
		}
		if(!success)
			pFilter->pReader->Release();
	}
	if(!success)
		CoUninitialize();
	return success;
}

long ReadWMAInput(FILTER_INPUT* pFilter, BYTE* buf, long lBytes)
{
	float* f=(float*)buf;
	long i;
	long a=0;
	while(a<lBytes)
	{
		if(pFilter->dwPos<pFilter->dwSize)
		{
			if(pFilter->bps==32)
			{
				i=*(pFilter->pData+pFilter->dwPos+2);
				i<<=8;
				i+=*(pFilter->pData+pFilter->dwPos+1);
				i<<=8;
				i+=*(pFilter->pData+pFilter->dwPos);
				i<<=8;
				*(f+a/4)=(float)i;
				*(f+a/4)/=65536.0f;
				pFilter->dwPos+=3;
				a+=4;
			}
			else
				*(buf+a++)=*(pFilter->pData+pFilter->dwPos++);
		}
		else
		{
			if(pFilter->pBuffer)
			{
				pFilter->pBuffer->Release();
				pFilter->pBuffer=0;
			}
			pFilter->pData=0;
			pFilter->dwSize=0;
			pFilter->dwPos=0;
			QWORD cnsSampleTime,cnsDuration;
			DWORD dwFlags,dwOutput;
			HRESULT hr=pFilter->pReader->GetNextSample(0,&pFilter->pBuffer,&cnsSampleTime,&cnsDuration,&dwFlags,&dwOutput,0);
			if(SUCCEEDED(hr))
				pFilter->pBuffer->GetBufferAndLength(&pFilter->pData,&pFilter->dwSize);
			else
				break;
		}
	}
	return a;
}

void CloseWMAInput(FILTER_INPUT* pFilter)
{
	if(pFilter->pBuffer)
		pFilter->pBuffer->Release();
	if(pFilter->pReader)
	{
		pFilter->pReader->Close();
		pFilter->pReader->Release();
	}
	CoUninitialize();
}

BOOL OpenWMAOutput(FILTER_OUTPUT* pFilter)
{
	HRESULT hr=CoInitializeEx(0,COINIT_APARTMENTTHREADED);
	if(FAILED(hr))
		return 0;
	BOOL success=0;
	hr=WMCreateWriter(0,&pFilter->pWriter);
	if(SUCCEEDED(hr))
	{
		IWMProfileManager* pProfileManager;
		hr=WMCreateProfileManager(&pProfileManager);
		if(SUCCEEDED(hr))
		{
			IWMProfile* pProfile;
			hr=pProfileManager->CreateEmptyProfile(WMT_VER_9_0,&pProfile);
			if(SUCCEEDED(hr))
			{
				IWMCodecInfo3* pCodecInfo;
				hr=pProfileManager->QueryInterface(IID_IWMCodecInfo3,(void**)&pCodecInfo);
				if(SUCCEEDED(hr))
				{
					if(!pFilter->dwOptions)
						pFilter->dwOptions=GetDefaultSetting();
					DWORD codecindex=(pFilter->dwOptions>>16)&0xff;
					DWORD formatindex=pFilter->dwOptions&0xffff;
					WMT_ATTR_DATATYPE datatype=WMT_TYPE_BOOL;
					BOOL vbr=0;
					if(pFilter->dwOptions&FLAG_VBR)
						vbr=1;
					hr=pCodecInfo->SetCodecEnumerationSetting(WMMEDIATYPE_Audio,codecindex,g_wszVBREnabled,datatype,(LPBYTE)&vbr,sizeof(BOOL));
					if(SUCCEEDED(hr))
					{
						IWMStreamConfig* pStreamConfig;
						hr=pCodecInfo->GetCodecFormat(WMMEDIATYPE_Audio,codecindex,formatindex,&pStreamConfig);
						if(SUCCEEDED(hr))
						{
							hr=pStreamConfig->SetStreamNumber(1);
							if(SUCCEEDED(hr))
							{
								hr=pProfile->AddStream(pStreamConfig);
								if(SUCCEEDED(hr))
								{
									hr=pFilter->pWriter->SetProfile(pProfile);
									if(SUCCEEDED(hr))
									{
										IWMInputMediaProps* pProps;
										hr=pFilter->pWriter->GetInputProps(0,&pProps);
										if(SUCCEEDED(hr))
										{
											DWORD dwSize=0;
											hr=pProps->GetMediaType(0,&dwSize);
											if(SUCCEEDED(hr)&&dwSize)
											{
												BYTE* pInput=new BYTE[dwSize];
												if(pInput)
												{
													hr=pProps->GetMediaType((WM_MEDIA_TYPE*)pInput,&dwSize);
													if(SUCCEEDED(hr))
													{
														WM_MEDIA_TYPE* pMediaType=(WM_MEDIA_TYPE*)pInput;
														WAVEFORMATEX* pwfx=(WAVEFORMATEX*)pMediaType->pbFormat;
														pwfx->cbSize=sizeof(WAVEFORMATEX);
														pwfx->wFormatTag=WAVE_FORMAT_PCM;
														pwfx->nChannels=pFilter->channels;
														pwfx->nSamplesPerSec=pFilter->samplerate;
														pwfx->wBitsPerSample=pFilter->bps;
														pwfx->nBlockAlign=pwfx->nChannels*(pwfx->wBitsPerSample/8);
														pwfx->nAvgBytesPerSec=pwfx->nSamplesPerSec*pwfx->nBlockAlign;
														hr=pProps->SetMediaType(pMediaType);
														if(SUCCEEDED(hr))
														{
															hr=pFilter->pWriter->SetInputProps(0,pProps);
															if(SUCCEEDED(hr))
															{
																WCHAR output[MAX_PATH];
																MultiByteToWideChar(CP_ACP,MB_PRECOMPOSED,pFilter->filename,-1,output,MAX_PATH);
																hr=pFilter->pWriter->SetOutputFilename(output);
																if(SUCCEEDED(hr))
																	success=1;
															}
														}
													}
													delete [] pInput;
												}
											}
											pProps->Release();
										}
									}
								}
							}
							pStreamConfig->Release();
						}
					}
					pCodecInfo->Release();
				}
				pProfile->Release();
			}
			pProfileManager->Release();
		}
		if(!success)
			pFilter->pWriter->Release();
	}
	if(!success)
		CoUninitialize();
	return success;
}

long WriteWMAOutput(FILTER_OUTPUT* pFilter, BYTE* pBuf, long lBytes)
{
	if(!pFilter->beginwriting)
	{
		if(SUCCEEDED(pFilter->pWriter->BeginWriting()))
			pFilter->beginwriting=1;
		else
			return 0;
	}
	long i=lBytes;
	if(pFilter->bps==24)
	{
		float* fBuf=(float*)pBuf;
		int a;
		for(a=0;a<lBytes/4;a++)
		{
			i=(long)(fBuf[a]*256);
			if(i>8388607)
				i=8388607;
			else
			{
				if(i<-8388607)
					i=-8388607;
			}
			pBuf[a*3]=(BYTE)(i&0xff);
			pBuf[a*3+1]=(BYTE)(i>>8&0xff);
			pBuf[a*3+2]=(BYTE)(i>>16&0xff);
		}
		i=a*3;
	}
	INSSBuffer* pBuffer=0;
	HRESULT	hr=pFilter->pWriter->AllocateSample(i,&pBuffer);
	if(SUCCEEDED(hr))
	{
		BYTE* pData;
		DWORD cbData;
		hr=pBuffer->GetBufferAndLength(&pData,&cbData);
		if(SUCCEEDED(hr))
		{
			memcpy(pData,pBuf,i);
			hr=pBuffer->SetLength(i);
			if(SUCCEEDED(hr))
			{
				hr=pFilter->pWriter->WriteSample(0,pFilter->cnsCurrent,0,pBuffer);
				QWORD qwTemp=cbData;
				qwTemp*=10000000;
				qwTemp/=(pFilter->samplerate*pFilter->channels*(pFilter->bps/8));
				pFilter->cnsCurrent+=qwTemp;
				pBuffer->Release();
				return lBytes;
			}
		}
		pBuffer->Release();
	}
	return 0;
}

void CloseWMAOutput(FILTER_OUTPUT* pFilter)
{
	if(pFilter->pWriter)
	{
		if(pFilter->beginwriting)
			pFilter->pWriter->EndWriting();
		pFilter->pWriter->Release();
	}
	CoUninitialize();
}

BOOL GetNextComment(FILTER_INPUT* pFilter, SPECIALDATA* psp)
{
	BOOL success=0;
	IWMHeaderInfo3* pHeader;
	HRESULT hr=pFilter->pReader->QueryInterface(IID_IWMHeaderInfo3,(void**)&pHeader);
	if(SUCCEEDED(hr))
	{
		UINT i=(UINT)(void*)psp->hSpecialData;
		WMT_ATTR_DATATYPE wmType;
		WORD langIndex;
		WORD wField;
		WCHAR* pField;
		DWORD dwValue;
		BYTE* pValue;
		while(SUCCEEDED(pHeader->GetAttributeByIndexEx(0,i,0,&wField,&wmType,&langIndex,0,&dwValue)))
		{
			pField=new WCHAR[wField];
			if(pField)
			{
				pValue=new BYTE[dwValue];
				if(pValue)
				{
					hr=pHeader->GetAttributeByIndexEx(0,i,pField,&wField,&wmType,&langIndex,pValue,&dwValue);
					if(SUCCEEDED(hr))
					{
						for(int j=0;j<TAGCOUNT;j++)
						{
							if(!_wcsicmp(pField,TAGS[j].tag))
							{
								char* ansi;
								if(UNICODEtoANSI((WCHAR*)pValue,&ansi))
								{
									int x=strlen(ansi)+2;
									BOOL bDisp=0;
									if(!strcmp(TAGS[j].riff,FCC_DISP))
									{
										bDisp=1;
										x+=4;
									}
									psp->hData=GlobalAlloc(GMEM_MOVEABLE|GMEM_ZEROINIT,x);
									if(psp->hData)
									{
										char* pData=(char*)GlobalLock(psp->hData);
										psp->dwSize=x-1;
										psp->dwExtra=1;
										if(bDisp)
										{
											strcpy(psp->szListType,FCC_WAVE);
											strcpy(pData+4,ansi);
											pData[0]=1;
										}
										else
										{
											strcpy(psp->szListType,FCC_INFO);
											strcpy(pData,ansi);
										}
										strcpy(psp->szType,TAGS[j].riff);
										GlobalUnlock(pData);
										success=1;
									}
									delete [] ansi;
								}
								break;
							}
						}
					}
					delete [] pValue;
				}
				delete [] pField;
			}
			psp->hSpecialData=(void*)(++i);
			if(success)
				break;
		}
		pHeader->Release();
	}
	return success;
}

BOOL SetComment(FILTER_OUTPUT* pFilter, WCHAR* field, WCHAR* value)
{
	BOOL success=0;
	IWMHeaderInfo* pHeader=0;
	if(SUCCEEDED(pFilter->pWriter->QueryInterface(IID_IWMHeaderInfo,(void**)&pHeader)))
	{
		if(SUCCEEDED(pHeader->SetAttribute(0,field,WMT_TYPE_STRING,(LPBYTE)value,(wcslen(value)+1)*sizeof(WCHAR))))
			success=1;
		pHeader->Release();
	}
	return success;
}

DWORD GetDefaultSetting()
{
	DWORD setting=0;
	HRESULT hr=CoInitializeEx(0,COINIT_APARTMENTTHREADED);
	if(SUCCEEDED(hr))
	{
		IWMProfileManager* pProfileManager;
		hr=WMCreateProfileManager(&pProfileManager);
		if(SUCCEEDED(hr))
		{
			IWMCodecInfo3* pCodecInfo;
			hr=pProfileManager->QueryInterface(IID_IWMCodecInfo3,(void**)&pCodecInfo);
			if(SUCCEEDED(hr))
			{
				IWMStreamConfig* pStreamConfig=0;
				IWMMediaProps* pMediaProps=0;
				BOOL vbr=0;
				DWORD cCodecs=0;
				DWORD cFormats=0;
				WCHAR* pwszFormatName=0;
				DWORD cchFormatName=0;
				DWORD dwSize=0;
				hr=pCodecInfo->GetCodecInfoCount(WMMEDIATYPE_Audio,&cCodecs);
				if(SUCCEEDED(hr))
				{
					WMT_ATTR_DATATYPE datatype=WMT_TYPE_BOOL;
					for(DWORD dwCodecIndex=0;dwCodecIndex<cCodecs;dwCodecIndex++)
				    {
						hr=pCodecInfo->SetCodecEnumerationSetting(WMMEDIATYPE_Audio,dwCodecIndex,g_wszVBREnabled,datatype,(LPBYTE)&vbr,sizeof(BOOL));
						if(SUCCEEDED(hr))
						{
							hr=pCodecInfo->GetCodecFormatCount(WMMEDIATYPE_Audio,dwCodecIndex,&cFormats);
							if(SUCCEEDED(hr)&&cFormats)
							{
								for(DWORD dwFormatIndex=0;dwFormatIndex<cFormats;dwFormatIndex++)
								{
									hr=pCodecInfo->GetCodecFormatDesc(WMMEDIATYPE_Audio,dwCodecIndex,dwFormatIndex,0,0,&cchFormatName);
									if(SUCCEEDED(hr))
									{
										pwszFormatName=new WCHAR[cchFormatName];
										if(pwszFormatName)
										{
											hr=pCodecInfo->GetCodecFormatDesc(WMMEDIATYPE_Audio,dwCodecIndex,dwFormatIndex,&pStreamConfig,pwszFormatName,&cchFormatName);
											if(SUCCEEDED(hr))
											{
												hr=pStreamConfig->QueryInterface(IID_IWMMediaProps,(void**)&pMediaProps);
												if(SUCCEEDED(hr))
												{
													hr=pMediaProps->GetMediaType(0,&dwSize);
													if(SUCCEEDED(hr)&&dwSize)
													{
														BYTE* pType=new BYTE[dwSize];
														if(pType)
														{
															hr=pMediaProps->GetMediaType((WM_MEDIA_TYPE*)pType,&dwSize);
															if(SUCCEEDED(hr))
															{
																WM_MEDIA_TYPE* pMediaType=(WM_MEDIA_TYPE*)pType;
																WAVEFORMATEX* pwfx=(WAVEFORMATEX*)pMediaType->pbFormat;
																if((pwfx->nChannels<=2)
																	&&(pwfx->wBitsPerSample==16)
																	&&(pwfx->nSamplesPerSec==44100)
																	&&(wcsstr(pwszFormatName,L"128 kbps"))
																	&&(!wcsstr(pwszFormatName,L"(A/V)")))
																{
																	setting=dwCodecIndex;
																	setting<<=16;
																	setting+=dwFormatIndex;
																}
															}
															delete [] pType;
														}
													}
													pMediaProps->Release();
												}
												pStreamConfig->Release();
											}
											delete [] pwszFormatName;
										}
									}
									if(setting)
										break;
								}
							}
						}
						if(setting)
							break;
					}
				}
				pCodecInfo->Release();
			}
			pProfileManager->Release();
		}
		CoUninitialize();
	}
	return setting;
}

void UpdateFormatCombo(HWND hWndDlg, BOOL vbr, DWORD codecindex)
{
	HRESULT hr=CoInitializeEx(0,COINIT_APARTMENTTHREADED);
	if(FAILED(hr))
		return;
	SendDlgItemMessage(hWndDlg,IDC_FORMAT,CB_RESETCONTENT,0,0);
	IWMProfileManager* pProfileManager;
	hr=WMCreateProfileManager(&pProfileManager);
	if(SUCCEEDED(hr))
	{
		IWMCodecInfo3* pCodecInfo;
		hr=pProfileManager->QueryInterface(IID_IWMCodecInfo3,(void**)&pCodecInfo);
		if(SUCCEEDED(hr))
		{
			IWMMediaProps* pMediaProps;
			IWMStreamConfig* pStreamConfig;
			WMT_ATTR_DATATYPE datatype=WMT_TYPE_BOOL;
			WCHAR* pwszFormatName=0;
			DWORD cchFormatName=0;
			char* pszCombo=0;
			DWORD cFormats=0;
			DWORD cCombo;
			hr=pCodecInfo->SetCodecEnumerationSetting(WMMEDIATYPE_Audio,codecindex,g_wszVBREnabled,datatype,(LPBYTE)&vbr,sizeof(BOOL));
			if(SUCCEEDED(hr))
			{
				hr=pCodecInfo->GetCodecFormatCount(WMMEDIATYPE_Audio,codecindex,&cFormats);
				if(SUCCEEDED(hr)&&cFormats)
				{
					for(DWORD index=0;index<cFormats;index++)
					{
						hr=pCodecInfo->GetCodecFormatDesc(WMMEDIATYPE_Audio,codecindex,index,0,0,&cchFormatName);
						if(SUCCEEDED(hr))
						{
							pwszFormatName=new WCHAR[cchFormatName];
							if(pwszFormatName)
							{
								hr=pCodecInfo->GetCodecFormatDesc(WMMEDIATYPE_Audio,codecindex,index,&pStreamConfig,pwszFormatName,&cchFormatName);
								if(SUCCEEDED(hr))
								{
									hr=pStreamConfig->QueryInterface(IID_IWMMediaProps,(void**)&pMediaProps);
									if(SUCCEEDED(hr))
									{
										hr=pMediaProps->GetMediaType(0,&cCombo);
										if(SUCCEEDED(hr)&&cCombo)
										{
											BYTE* pType=new BYTE[cCombo];
											if(pType)
											{
												hr=pMediaProps->GetMediaType((WM_MEDIA_TYPE*)pType,&cCombo);
												if(SUCCEEDED(hr))
												{
													WM_MEDIA_TYPE* pMediaType=(WM_MEDIA_TYPE*)pType;
													WAVEFORMATEX* pwfx=(WAVEFORMATEX*)pMediaType->pbFormat;
													if((pwfx->nChannels<=2)&&(!wcsstr(pwszFormatName,L"(A/V)")))
													{
														cCombo=WideCharToMultiByte(CP_ACP,0,pwszFormatName,cchFormatName,0,0,0,0);
														pszCombo=new char[cCombo];
														if(pszCombo)
														{
															WideCharToMultiByte(CP_ACP,0,pwszFormatName,cchFormatName,pszCombo,cCombo,0,0);
															cCombo=SendDlgItemMessage(hWndDlg,IDC_FORMAT,CB_ADDSTRING,0,(LPARAM)pszCombo);
															SendDlgItemMessage(hWndDlg,IDC_FORMAT,CB_SETITEMDATA,cCombo,index);
															delete [] pszCombo;
														}
													}
												}
												delete [] pType;
											}
										}
										pMediaProps->Release();
									}
									pStreamConfig->Release();
								}
								delete [] pwszFormatName;
							}
						}
					}
				}
			}
			pCodecInfo->Release();
		}
		pProfileManager->Release();
	}
	SendDlgItemMessage(hWndDlg,IDC_FORMAT,CB_SETCURSEL,0,0);
	CoUninitialize();
}

void UpdateCodecCombo(HWND hWndDlg, BOOL vbr, DWORD codecindex)
{
	HRESULT hr=CoInitializeEx(0,COINIT_APARTMENTTHREADED);
	if(FAILED(hr))
		return;
	SendDlgItemMessage(hWndDlg,IDC_CODEC,CB_RESETCONTENT,0,0);
	IWMProfileManager* pProfileManager;
	hr=WMCreateProfileManager(&pProfileManager);
	if(SUCCEEDED(hr))
	{
		IWMCodecInfo3* pCodecInfo;
		hr=pProfileManager->QueryInterface(IID_IWMCodecInfo3,(void**)&pCodecInfo);
		if(SUCCEEDED(hr))
		{
			DWORD cCodecs=0;
			WCHAR* pwszCodecName=0;
			DWORD cchCodecName=0;
			char* pszCombo=0;
			DWORD cCombo;
			hr=pCodecInfo->GetCodecInfoCount(WMMEDIATYPE_Audio,&cCodecs);
			if(SUCCEEDED(hr))
			{
				WMT_ATTR_DATATYPE datatype=WMT_TYPE_BOOL;
				DWORD cFormats=0;
				for(DWORD dwCodecIndex=0;dwCodecIndex<cCodecs;dwCodecIndex++)
			    {
					hr=pCodecInfo->SetCodecEnumerationSetting(WMMEDIATYPE_Audio,dwCodecIndex,g_wszVBREnabled,datatype,(LPBYTE)&vbr,sizeof(BOOL));
					if(SUCCEEDED(hr))
					{
						hr=pCodecInfo->GetCodecFormatCount(WMMEDIATYPE_Audio,dwCodecIndex,&cFormats);
						if(SUCCEEDED(hr)&&cFormats)
						{
							hr=pCodecInfo->GetCodecName(WMMEDIATYPE_Audio,dwCodecIndex,0,&cchCodecName);
							if(SUCCEEDED(hr))
							{
								pwszCodecName=new WCHAR[cchCodecName];
								if(pwszCodecName)
								{
									hr=pCodecInfo->GetCodecName(WMMEDIATYPE_Audio,dwCodecIndex,pwszCodecName,&cchCodecName);
									if(SUCCEEDED(hr))
									{
										cCombo=WideCharToMultiByte(CP_ACP,0,pwszCodecName,cchCodecName,0,0,0,0);
										pszCombo=new char[cCombo];
										if(pszCombo)
										{
											WideCharToMultiByte(CP_ACP,0,pwszCodecName,cchCodecName,pszCombo,cCombo,0,0);
											cCombo=SendDlgItemMessage(hWndDlg,IDC_CODEC,CB_ADDSTRING,0,(LPARAM)pszCombo);
											SendDlgItemMessage(hWndDlg,IDC_CODEC,CB_SETITEMDATA,cCombo,dwCodecIndex);
											if(dwCodecIndex==codecindex)
												SendDlgItemMessage(hWndDlg,IDC_CODEC,CB_SETCURSEL,cCombo,0);
											delete [] pszCombo;
										}
									}
									delete [] pwszCodecName;
								}
							}
						}
					}
				}
			}
			pCodecInfo->Release();
		}
		pProfileManager->Release();
	}
	if(SendDlgItemMessage(hWndDlg,IDC_CODEC,CB_GETCURSEL,0,0)==CB_ERR)
		SendDlgItemMessage(hWndDlg,IDC_CODEC,CB_SETCURSEL,0,0);
	CoUninitialize();
	UpdateFormatCombo(hWndDlg,vbr,SendDlgItemMessage(hWndDlg,IDC_CODEC,CB_GETITEMDATA,SendDlgItemMessage(hWndDlg,IDC_CODEC,CB_GETCURSEL,0,0),0));
}

BOOL ANSItoUNICODE(char* ansi, WCHAR** unicode)
{
	BOOL success=0;
	int i=MultiByteToWideChar(CP_ACP,MB_PRECOMPOSED,ansi,-1,0,0);
	if(i)
	{
		*unicode=new WCHAR[i];
		if(*unicode)
		{
			MultiByteToWideChar(CP_ACP,MB_PRECOMPOSED,ansi,-1,*unicode,i);
			success=1;
		}
	}
	return success;
}

BOOL UNICODEtoANSI(WCHAR* unicode, char** ansi)
{
	BOOL success=0;
	WCHAR* w=new WCHAR[wcslen(unicode)+1];
	if(w)
	{
		wcscpy(w,unicode);
		EnsureCRLF(&w);
		int i=WideCharToMultiByte(CP_ACP,0,w,-1,0,0,0,0);
		if(i)
		{
			*ansi=new char[i];
			if(*ansi)
			{
				WideCharToMultiByte(CP_ACP,0,w,-1,*ansi,i,0,0);
				success=1;
			}
		}
		delete [] w;
	}
	return success;
}

void EnsureCRLF(WCHAR** str)
{
	WCHAR* a=*str;
	WCHAR* b;
	int i;
	while(b=wcsstr(a,L"\r\n"))
	{
		a=++b+1;
		i=wcslen(a);
		if(i)
			memmove(b,a,i*2);
		*(b+i)=0;
	}
	a=*str;
	i=0;
	while(b=wcspbrk(a,L"\r\n"))
	{
		a=b+1;
		i++;
	}
	if(i)
	{
		a=*str;
		i=wcslen(a)+i+1;
		WCHAR* w=new WCHAR[i];
		if(w)
		{
			ZeroMemory(w,i*sizeof(WCHAR));
			while(b=wcspbrk(a,L"\r\n"))
			{
				wcsncat(w,a,b-a);
				wcscat(w,L"\r\n");
				a=b+1;
			}
			wcscat(w,a);
			delete [] *str;
			*str=w;
		}
	}
}
